/*===============================================*/
/* 1-Pass-Compiler Infix -> Postfix              */
/*===============================================*/

#include <stdio.h>
#include <ctype.h>

/*--- Token types ---*/

#define NONE  -1

#define NUM   256
#define DIV   257
#define MOD   258
#define ID    259
#define SEM   260
#define DONE  261

/*--- Symbol table ---*/

/* Size of table */
#define SYMMAX 100 

/* Table entry */
struct entry {
     char *lexptr;
     int  token ;
};

/* Table */
struct entry symtable [SYMMAX];

/* Last used table entry */
int  lastentry;

/* Reserved words */
struct entry keywords[] = {
  "div",DIV,
  "mod",MOD,
  0,0
};

/*--- Lexem array ---*/

/* End of string */ 
#define EOS   '\0'

/* Size of lexem array */
#define STRMAX 999

/* Lexem array */
char lexemes[STRMAX];

/* Last used lexem position */
int  lastchar;    

/*--- Token buffer ---*/

/* Size of buffer */
#define BSIZE 128

/* Buffer */
char lexbuf[BSIZE];

/*-----------------------------------------------*/
/* Management of symbol table                    */
/*-----------------------------------------------*/

/* Initialize symbol table */
init()  
{
  struct entry *p;

  lastentry = 0;
  lastchar = -1;
  for (p=keywords; p->token; p++)
          insert(p->lexptr, p->token);
}/*init*/

/* Check whether symbol is in table */
int  
lookup(char s[])    
{
  int p;

  for (p=lastentry; p>0; p=p-1) {
    if (strcmp(symtable[p].lexptr,s)==0) {
      return p;
    }
  }

  return 0;
}/*lookup*/

/* Insert new symbol into table */
int  
insert(char s[], int tok) 
{
  int len;

  len = strlen(s);

  if (lastentry+1>=SYMMAX) {
    error ("Symbol table overflow!");
  }

  if (lastchar+len+1>=STRMAX) {
    error ("Lexem array overflow!");
  }

  lastentry++;
  symtable[lastentry].token = tok;
  symtable[lastentry].lexptr = &lexemes[lastchar+1];
  lastchar = lastchar + len + 1;
  strcpy(symtable[lastentry].lexptr, s);

  return lastentry;
}/*insert*/

/* Lookahead symbol */
int lookahead;

/*-----------------------------------------------*/
/* Scanner                                       */
/*-----------------------------------------------*/

/* Line number */
int lineno;

/* Token value */
int tokenval;

/* Get next token */
int lexan()
{
  int t;

  while(1) {
    t = getchar();
    if (t==' ' ||t=='\t') {
      /* Eliminate blanks */
      ;
    }
    else if (t=='\n') {
      /* Increment linenumber */
      lineno++;
    }
    else if (isdigit(t)) {
      /* t is digit */
      ungetc(t,stdin);
      scanf("%d", &tokenval);
      return NUM;
    }
    else if (isalpha(t)) {     
      /* t is letter */
      int p, b = 0;

      while (isalnum(t)) {  /* t is alphanumeric */
        lexbuf[b] = t;
        t = getchar();
        b++;
        if (b >= BSIZE) {
          error("Compiler error!");
        }
      }

      lexbuf[b]= EOS;
      if (t != EOF) { 
        ungetc(t,stdin);
      }
      p = lookup(lexbuf);
      if (p==0) {
        p = insert (lexbuf, ID);
      }
      tokenval = p;
      return symtable[p].token;
    }
    else if (t==EOF) {
      /* End of file */
      return DONE;
    }
    else if (t==';') {
      /* End of statement */
      tokenval= SEM;
      return t;
    }
    else {
      /* No token recognized */
      tokenval= NONE;
      return t;
    }
  }
}/*lexan*/

/* Match token and get next one */
match(int t)
{
  if (lookahead==t) {
    lookahead =lexan();
  }
  else 
    error ("Syntax error");
}/*match*/

/*-----------------------------------------------*/
/* Parser                                        */
/*-----------------------------------------------*/

/* Parse input */
parse()            
{
  lookahead = lexan();
  while (lookahead!=DONE) {
    expr(); 
    match (';');
    printf("\n"); 
  }

}

/* Parse nonterminal expr */
expr()
{
  int t;

  term();
  while(1) {
    switch(lookahead) {
    case '+': 
    case '-':
      t = lookahead;
      match(lookahead); 
      term(); 
      emit(t,NONE);
      break;
    default:
      return;
    }
  }
}/*expr*/

/* Parse nonterminal term */
term()
{
  int t;

  factor();
  while(1) {
    switch(lookahead) {
      case '*': 
      case '/': 
      case DIV: 
      case MOD:
        t = lookahead;
        match(lookahead); 
        factor(); 
        emit(t,NONE);
        break;
      default:
        return;
    }
  }
}/*term*/

/* Parse nonterminal factor */
factor()
{
  switch(lookahead) {
  case '(':
    match ('('); 
    expr(); 
    match (')'); 
    break;
  case NUM:
    emit(NUM,tokenval);
    match(NUM); 
    break;
  case ID:
    emit(ID,tokenval);
    match(ID); 
    break;
  default:
    error("Syntax error");
  }

}

/*-----------------------------------------------*/
/* Compiler output                               */
/*-----------------------------------------------*/

emit(int t, int tval)      
{
  switch(t) {
  case '+': 
  case '-': 
  case '*': 
  case '/':
    printf("%c ", t); 
    break;
  case DIV:
    printf("DIV "); 
    break;
  case MOD:
    printf("MOD "); 
    break;
  case NUM:
    printf("%d ", tval); 
    break;
  case ID:
    printf("%s ", symtable[tval].lexptr); 
    break;
  default:
    printf("token %d, tokenval %d ", t, tval);
    break;
  }
}/*emit*/

/*-----------------------------------------------*/
/* Error messages                                */
/*-----------------------------------------------*/

error(char *m)           
{

  printf("ERROR: line %d:%s\n", lineno, m);
  printf("\nInterrupted.\n");
  exit(1); 

}

/*-----------------------------------------------*/
/* Main program                                  */
/*-----------------------------------------------*/

main()
{
  lineno = 1;
  tokenval = NONE;

  printf("\nInfix-Postfix-Compiler\n----------------------\n\n");

  init();
  parse();

  printf("\nTerminated.\n");
  exit(0);

}
